#!/usr/bin/env python

"""
Preparation of run-time attribute lookup tables.
"""

try:
    set
except NameError:
    from sets import Set as set

class List:

    """
    A displaced list containing attribute details and an index mapping names to
    offsets in the list.
    """

    def __init__(self, displaced, offsets, names):
        self.displaced = displaced
        self.offsets = offsets
        self.names = names
        self.names_index = {}
        for i, name in enumerate(self.names):
            self.names_index[name] = i

    def access(self, objname, attrname):
        obj_offset = self.offsets.get(objname)
        attr_index = self.names_index.get(attrname)
        if obj_offset is None or attr_index is None:
            return None

        element = self.displaced[obj_offset + attr_index]
        if element is not None:
            offset, details = element
            if offset == obj_offset:
                return details

        return None

class Table:

    "A lookup table."

    def __init__(self):
        self.attributes = set()
        self.table = {}
        self.names = []

    def add(self, objname, attributes):

        "For the given 'objname' add the given 'attributes' to the table."

        self.table[objname] = attributes
        for name, origin in attributes.items():
            self.attributes.add(name)

    def attribute_names(self):

        "Return the attribute names used in the table."

        self.names = self.names or list(self.attributes)
        return self.names

    def as_matrix(self):

        """
        Return the table as a matrix mapping object names to lists of attributes
        arranged in the order of the recorded attribute names.
        """

        matrix = {}
        for objname, attributes in self.table.items():
            row = []
            for name in self.attribute_names():
                row.append(attributes.get(name))
            matrix[objname] = row
        return matrix

    def as_list(self):

        """
        Return a displaced list object encoding the table in a more compact form
        than that provided by the matrix representation.
        """

        displaced = []
        offsets = {}
        offsets_used = set()

        # Visit each row of the matrix.

        for objname, attributes in self.as_matrix().items():
            len_displaced = len(displaced)
            len_attributes = len(attributes)

            # Try to fit the row into the list, starting from the beginning and
            # skipping places where an existing row has been fitted.

            for offset in xrange(0, len_displaced):
                if offset in offsets_used:
                    continue
                if self._fit_row(offset, attributes, len_attributes, displaced, len_displaced):
                    offsets_used.add(offset)
                    offsets[objname] = offset
                    self._add_row(offset, attributes, len_attributes, displaced, len_displaced)
                    break

            # Where a row could not be fitted within the list, add it to the
            # end.

            else:
                offsets_used.add(len_displaced)
                offsets[objname] = len_displaced
                self._add_row(len_displaced, attributes, len_attributes, displaced, len_displaced)

        return List(displaced, offsets, self.names)

    def _fit_row(self, offset, attributes, len_attributes, displaced, len_displaced):

        """
        Fit, at the given 'offset', the row of 'attributes' having length
        'len_attributes' in the 'displaced' list having length 'len_displaced'.
        Return a true value if this succeeded.
        """

        for i, attr in enumerate(attributes):
            if i + offset >= len_displaced:
                break
            element = displaced[i + offset]
            if attr is not None and element is not None:
                return 0
        return 1

    def _add_row(self, offset, attributes, len_attributes, displaced, len_displaced):

        """
        Add, at the given 'offset', the row of 'attributes' having length
        'len_attributes' in the 'displaced' list having length 'len_displaced'.
        """

        # Extend the list if necessary.

        for i in xrange(0, offset + len_attributes - len_displaced):
            displaced.append(None)

        # Record the offset and attribute details in the list.

        for i, attr in enumerate(attributes):
            if attr is not None:
                displaced[offset+i] = offset, attr

# vim: tabstop=4 expandtab shiftwidth=4
